#    Copyright 2025 Genesis Corporation.
#
#    All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


RA_SERVICE_NAME = "ra_test"


RA_POSTGRESQL_PROTO_NAME = "postgresql"
RA_POSTGRESQL_DB_PORT = 5432
RA_POSTGRESQL_DB_HOST = "127.0.0.1"
RA_POSTGRESQL_USERNAME = "postgres"
RA_POSTGRESQL_PASSWORD = "password"
RA_POSTGRESQL_DB_NAME = RA_SERVICE_NAME


RA_MYSQL_PROTO_NAME = "mysql"
RA_MYSQL_DB_PORT = 3306
RA_MYSQL_DB_HOST = "127.0.0.1"
RA_MYSQL_USERNAME = "root"
RA_MYSQL_PASSWORD = "password"
RA_MYSQL_DB_NAME = RA_SERVICE_NAME

DB_CONFIG_SECTION = "db"
